@@ -1,10 +1,12 @@
 #!/usr/bin/env python
 # -*- coding: utf8 -*-
 
+from pathlib import Path
 import numpy as np
 import pandas as pd
 from tqdm import tqdm
 import xml.etree.ElementTree as ET
+from lxml import etree
 import json
 import glob
 from mhealthdata.utils import *
@@ -31,17 +33,22 @@ class DataLoader():
     Methods are device-centric and can be used to get data for any specific device
     connected to your health app aggregater, e.g. iPhone, or Apple Watch, or Fitbit wristband
 
-    Note
-    ----
+    Notes
+    -----
     Other data can be found in data export (e.g. VO2Max), but not processed by DataLoader subclasses.
-    In case those data needed, the self.dtype dict attribute should be modified.
+    In case those data needed, the self.category dict attribute should be modified.
+
+    Parameters
+    ----------
+    path : str
+        Path to unzipped local folder containing health app data
 
     Attributes
     ----------
     df : dict
         Dictionary of pandas Dataframes of loaded health data for "steps", "bpm", etc.
-    dtypes : dict
-        Dictionary of health data to load. Keys used to find files. Attributes: 
+    categories : dict
+        Dictionary of health data categories. Keys used to find files. Attributes: 
         "name" - to rename, "column" - to seek value column in corresponding Dataframe.
     userdata : dict
         Dictionary of "Date-of-birth", "Biological sex", and "Height".
@@ -60,17 +67,8 @@ class DataLoader():
     """
 
     def __init__(self, path):
-        """
-        Initialize with loading data from an export folder path.
-
-        Parameters
-        ----------
-        path : str
-            Path to unzipped local folder containing health app data
-
-        """
         self.df = {}
-        self.dtypes = {}
+        self.categories = {}
         self.userdata = {}
         self.start_keys = ["start_time", "startTime", "startDate", 
                            "dateTime", "day_time"]
@@ -123,8 +121,23 @@ class DataLoader():
         return list(self.df.keys())
 
 
+    @property
+    def all_categories(self):
+        """
+        Get list of all data categories found (not all loaded).
+        
+        Returns
+        -------
+        list
+            List of all data categories found in provided path.
+
+        """
+        categories = []
+        return categories
+
+
     @staticmethod
-    def _special_cases(df, dtype):
+    def _special_cases(df, category):
         """
         Private method to process special cases during loading of data.
         
@@ -132,7 +145,7 @@ class DataLoader():
         ----------
         df : Dataframe
             Dataframes of loaded health data for "steps", "bpm", etc.
-        dtype : str
+        category : str
             Key used to find health data files.
 
         Returns
@@ -145,7 +158,7 @@ class DataLoader():
 
 
     @staticmethod
-    def _get_values(df, dtype):
+    def _get_values(df, category):
         """
         Private method to get record value arrays from loaded datframes.
         
@@ -153,7 +166,7 @@ class DataLoader():
         ----------
         df : Dataframe
             Dataframes of loaded health data for "steps", "bpm", etc.
-        dtype : str
+        category : str
             Key used to find health data files.
 
         Returns
@@ -163,8 +176,9 @@ class DataLoader():
 
         """
         values = df.values
-        if "sleep" in dtype.lower():
+        if "sleep" in category.lower():
             s = sleep_stage_dict(mode="encode")
+            s["None"] = 0
             values = np.vectorize(s.get)(values)
         values = values.astype(float)
         return values
@@ -423,13 +437,13 @@ class DataLoader():
         return df
 
         
-    def _parse_dataframe(self, dtype, device):
+    def _parse_dataframe(self, category, device):
         """
         Private method to extract record data from a (device-sliced) Dataframe.
         
         Parameters
         ----------
-        dtype : str
+        category : str
             Key used to find health data files.
         device : str
             Device name (sould match any one of self.devices).
@@ -447,14 +461,14 @@ class DataLoader():
 
         """
         value = idate = imin = duration = np.array([])
-        if dtype in self.df and device in self.devices:
-            df = self.df[dtype]
-            if dtype not in ["weight"] and device not in ["all"]:
+        if category in self.df and device in self.devices:
+            df = self.df[category]
+            if category not in ["weight"] and device not in ["all"]:
                 uuids = self.devices_dict[device]
                 df = self._get_device_slice(df, uuids, self.dev_col)
             if df.shape[0] > 0:
-                column = self.dtypes[dtype]["column"]
-                value = self._get_values(df[column], dtype)
+                column = self.categories[category]["column"]
+                value = self._get_values(df[column], category)
                 t0, t1 = [self._get_time(df, k) for k in [self.start_keys, self.end_keys]]
                 idate, imin = self._get_idate_imin(t0)
                 duration = self._get_duration(df)
@@ -530,9 +544,9 @@ class DataLoader():
         data = {}
         if device not in self.devices:
             raise KeyError(f"Wrong device '{device}'. Use 'devices' property to get valid devices.")
-        for dtype in self.dtypes:
-            dname = self.dtypes[dtype]["name"]
-            value, idate, imin, dt = self._parse_dataframe(dtype, device)
+        for category in self.categories:
+            dname = self.categories[category]["name"]
+            value, idate, imin, dt = self._parse_dataframe(category, device)
             try:
                 date_range = idate if date_range is None else date_range
                 date_range = dates_to_range(date_range)
@@ -590,8 +604,8 @@ class DataLoader():
 class FitbitLoader(DataLoader):
     """
     
-    Note
-    ----
+    Notes
+    -----
     One may note that Fitbit exported
     
         - ``sleep`` in local time
@@ -616,7 +630,7 @@ class FitbitLoader(DataLoader):
 
     def __init__(self, path):
         super().__init__(path)
-        self.dtypes = {
+        self.categories = {
             "steps": {
                 "name": "steps", "column": "value"},
             "sleep": {
@@ -631,9 +645,21 @@ class FitbitLoader(DataLoader):
         self.load_data()
 
 
+    @property
+    def all_categories(self):
+        fnames = glob.glob(self.path + "/*/*.csv")
+        fnames += glob.glob(self.path + "/*/*.json")
+        categories = []
+        for fname in fnames:
+            category = Path(fname).stem.split("-")[0]
+            if category not in categories:
+                categories.append(category)
+        return categories
+
+
     @staticmethod
-    def _special_cases(df, dtype):
-        if dtype == "weight":
+    def _special_cases(df, category):
+        if category == "weight":
             df["dateTime"] = df["date"] + " " + df["time"]
             df["weight"] = 0.4536 * df["weight"] # pounds to kg
         return df
@@ -662,7 +688,7 @@ class FitbitLoader(DataLoader):
     def load_sleep(self):
         """
         Load sleep data from .json files.
-        Peth to seek files is taken from self.path attribute.
+        Path to seek files is taken from self.path attribute.
         
         Returns
         -------
@@ -683,14 +709,14 @@ class FitbitLoader(DataLoader):
         return df
 
 
-    def load_nonsleep(self, dtype):
+    def load_nonsleep(self, category):
         """
         Load non-sleep data from .json files.
-        Peth to seek files is taken from self.path attribute.
+        Path to seek files is taken from self.path attribute.
         
         Parameters
         ----------
-        dtype : str
+        category : str
             Key used to find health data files.
 
         Returns
@@ -699,7 +725,7 @@ class FitbitLoader(DataLoader):
             Dataframe of raw data loaded from .csv.
 
         """
-        fnamelist = glob.glob(self.path + "/*/" + dtype + "-*")
+        fnamelist = glob.glob(self.path + "/*/" + category + "-*")
         if len(fnamelist) == 0:
             return None
         df = [pd.json_normalize(json.load(open(fname))) for fname in tqdm(fnamelist)]
@@ -710,8 +736,8 @@ class FitbitLoader(DataLoader):
     def load_data(self):
         """
         Load data from .csv and .json files.
-        Cycling over dtype from self.dtypes attribute.
-        Peth to seek files is taken from self.path attribute.
+        Cycling over category from self.categories attribute.
+        Path to seek files is taken from self.path attribute.
         
         Returns
         -------
@@ -720,14 +746,14 @@ class FitbitLoader(DataLoader):
 
         """
         self.df = {}
-        for dtype in list(self.dtypes.keys()):
-            df = self.load_sleep() if dtype == "sleep" else self.load_nonsleep(dtype)
+        for category in list(self.categories.keys()):
+            df = self.load_sleep() if category == "sleep" else self.load_nonsleep(category)
             if df is not None:
-                df = self._special_cases(df, dtype)
+                df = self._special_cases(df, category)
                 df = self._parse_timestamps(df)
-                self.df[dtype] = df
-            elif "step" in dtype or "pedometer" in dtype:
-                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{dtype}'.")
+                self.df[category] = df
+            elif "step" in category or "pedometer" in category:
+                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{category}'.")
         self._parse_userdata()
         return self.dataframes
 
@@ -742,8 +768,8 @@ class FitbitLoader(DataLoader):
 class ShealthLoader(DataLoader):
     """
     
-    Note
-    ----
+    Notes
+    -----
     One may note that Samsung Health exported
     
         - ``.json`` (``step`` binning data) timestamps in local time
@@ -771,7 +797,7 @@ class ShealthLoader(DataLoader):
         super().__init__(path)
         self.start_keys = ["start_time"]
         self.dev_col = ["deviceuuid"]
-        self.dtypes = {
+        self.categories = {
             "pedometer_day_summary": {
                 "name": "steps", "column": "mStepCount"},
             "step_daily_trend": {
@@ -816,9 +842,20 @@ class ShealthLoader(DataLoader):
         return dev
 
     
+    @property
+    def all_categories(self):
+        fnames = glob.glob(self.path + "/*/*.csv")
+        categories = []
+        for fname in fnames:
+            category = fname.split(".")
+            if len(category) > 3:
+                categories.append(category)
+        return categories
+
+
     @staticmethod
-    def _special_cases(df, dtype):
-        if "sleep" in dtype:
+    def _special_cases(df, category):
+        if "sleep" in category:
             if "stage" in df.columns:
                 s = {40001: "awake", 40002: "light", 40003: "deep", 40004: "rem"}
                 df["stage"] = np.vectorize(s.get)(df["stage"].values)
@@ -839,13 +876,13 @@ class ShealthLoader(DataLoader):
         return list(self.userdata.keys())
 
 
-    def _binning_dict(self, dtype, idx):
+    def _binning_dict(self, category, idx):
         """
         Private method to get device-id and date dictionaries for daily .json.
         
         Parameters
         ----------
-        dtype : str
+        category : str
             Key used to find health data files.
         idx : str
             Column name containing binning data file names.
@@ -858,21 +895,21 @@ class ShealthLoader(DataLoader):
             Dictionary of dates for binning data file names.
 
         """
-        df = self.load_csv(dtype)
+        df = self.load_csv(category)
         df.set_index(idx, inplace=True)
         dat = df["day_time"].to_dict()
         dev = df["deviceuuid"].to_dict()
         return dev, dat
 
 
-    def load_csv(self, dtype):
+    def load_csv(self, category):
         """
         Load data from .csv file.
-        Peth to seek files is taken from self.path attribute.
+        Path to seek files is taken from self.path attribute.
         
         Parameters
         ----------
-        dtype : str
+        category : str
             Key used to find health data files.
 
         Returns
@@ -882,21 +919,21 @@ class ShealthLoader(DataLoader):
 
         """
         try:
-            fname = glob.glob(self.path + "/*/*." + dtype + ".*.csv")[0]
+            fname = glob.glob(self.path + "/*/*." + category + ".*.csv")[0]
         except IndexError as e:
             return None
         df = pd.read_csv(fname, skiprows=1, index_col=False)
         return df
 
 
-    def load_jsons(self, dtype, idx="binning_data"):
+    def load_jsons(self, category, idx="binning_data"):
         """
         Load data from .json files.
-        Peth to seek files is taken from self.path attribute.
+        Path to seek files is taken from self.path attribute.
         
         Parameters
         ----------
-        dtype : str
+        category : str
             Key used to find health data files.
         idx : str, default "binning_data"
             Column name containing binning data file names.
@@ -907,8 +944,8 @@ class ShealthLoader(DataLoader):
             Dataframe of raw data loaded from .csv.
 
         """
-        dev, dat = self._binning_dict(dtype, idx)
-        fnamelist = glob.glob(self.path + '/*/jsons/*' + dtype + '*/*/*.' + idx + '.json')
+        dev, dat = self._binning_dict(category, idx)
+        fnamelist = glob.glob(self.path + '/*/jsons/*' + category + '*/*/*.' + idx + '.json')
         if len(fnamelist) == 0:
             return None
         df = []
@@ -928,8 +965,8 @@ class ShealthLoader(DataLoader):
     def load_data(self):
         """
         Load data from .csv and .json files.
-        Cycling over dtype from self.dtypes attribute.
-        Peth to seek files is taken from self.path attribute.
+        Cycling over category from self.categories attribute.
+        Path to seek files is taken from self.path attribute.
         
         Returns
         -------
@@ -938,16 +975,16 @@ class ShealthLoader(DataLoader):
 
         """
         self.df = {}
-        for dtype in list(self.dtypes.keys()) + ["device_profile", "user_profile"]:
-            df = self.load_csv(dtype)
+        for category in list(self.categories.keys()) + ["device_profile", "user_profile"]:
+            df = self.load_csv(category)
             if "binning_data" in df.columns:
-                df = self.load_jsons(dtype)
+                df = self.load_jsons(category)
             if df is not None:
-                df = self._special_cases(df, dtype)
+                df = self._special_cases(df, category)
                 df = self._parse_timestamps(df)
-                self.df[dtype] = df
-            elif "step" in dtype or "pedometer" in dtype:
-                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{dtype}'.")
+                self.df[category] = df
+            elif "step" in category or "pedometer" in category:
+                raise FileNotFoundError(f"Wrong 'path'. Cannot find files for '{category}'.")
         self._parse_userdata()
         return self.dataframes
 
@@ -979,7 +1016,7 @@ class HealthkitLoader(DataLoader):
     def __init__(self, path):
         super().__init__(path)
         self.dev_col = ["sourceName"]
-        self.dtypes = {
+        self.categories = {
             "HKQuantityTypeIdentifierStepCount": {
                 "name": "steps", "column": "value"},
             "HKCategoryTypeIdentifierSleepAnalysis": {
@@ -999,16 +1036,30 @@ class HealthkitLoader(DataLoader):
     @property
     def devices_dict(self):
         dev = {"all": ["all"]}
-        for dtype in self.df:
-            df = self.df[dtype]
-            if "sourceName" in df.columns:
-                for d in np.unique(df["sourceName"].values.astype(str)):
-                    dev[d] = [d]
+        for category in self.categories:
+            if category in self.df:
+                df = self.df[category]
+                if "sourceName" in df.columns:
+                    for d in np.unique(df["sourceName"].values.astype(str)):
+                        dev[d] = [d]
         return dev
 
 
+    @property
+    def all_categories(self):
+        categories = []
+        for tag in ["Record", "Workout"]:
+            df = self.df[tag]
+            col = find_columns_by_key(df, ["type"])
+            if len(col) > 0:
+                categories.append(df[col[0]].values.astype(str))
+        categories = np.concatenate(categories)
+        categories = np.unique(categories).tolist()
+        return categories
+
+
     @staticmethod
-    def _special_cases(df, dtype):
+    def _special_cases(df, category):
         def clean_username(s):
             if "iphone" in s.lower():
                 s = "iPhone"
@@ -1017,7 +1068,7 @@ class HealthkitLoader(DataLoader):
             return s
         if "sourceName" in df.columns:
             df["sourceName"] = df["sourceName"].apply(clean_username)
-        if dtype == "HKCategoryTypeIdentifierSleepAnalysis":
+        if category == "HKCategoryTypeIdentifierSleepAnalysis":
             s = {"HKCategoryValueSleepAnalysisAwake": "awake", 
                  "HKCategoryValueSleepAnalysisInBed": "no_stage", 
                  "HKCategoryValueSleepAnalysisAsleep": "asleep"}
@@ -1041,7 +1092,7 @@ class HealthkitLoader(DataLoader):
     def _parse_xml(self, root):
         """
         Private method to parse records from loaded .xml
-        Peth to seek files is taken from self.path attribute.
+        Path to seek files is taken from self.path attribute.
         
         Parameters
         ----------
@@ -1063,7 +1114,7 @@ class HealthkitLoader(DataLoader):
                         if node.tag == "MetadataEntry":
                             if node.attrib["key"] == "HKTimeZone":
                                 child.attrib["HKTimeZone"] = node.attrib["value"]
-                    records.append(child.attrib)
+                    records.append(dict(child.attrib))
             df = pd.DataFrame(records)
             df = self._parse_timestamps(df)
             data[tag] = df
@@ -1073,8 +1124,8 @@ class HealthkitLoader(DataLoader):
     def load_data(self):
         """
         Load data from .csv and .json files.
-        Cycling over dtype from self.dtypes attribute.
-        Peth to seek files is taken from self.path attribute.
+        Cycling over category from self.categories attribute.
+        Path to seek files is taken from self.path attribute.
         
         Returns
         -------
@@ -1086,15 +1137,18 @@ class HealthkitLoader(DataLoader):
             fname = glob.glob(self.path + "/[eE][xX][pP][oO][rR][tT].[xX][mM][lL]")[0]
         except IndexError as e:
             raise FileNotFoundError(f"Wrong 'path'. Cannot find file 'export.xml'.")
-        tree = ET.parse(fname)
+        parser = etree.XMLParser(recover=True)
+        tree = ET.parse(fname, parser=parser)
         root = tree.getroot()
         data = self._parse_xml(root)
+        for tag in data:
+            self.df[tag] = data[tag]
         self._parse_userdata(data)
-        for dtype in self.dtypes:
-            mask = data["Record"]["type"] == dtype
+        for category in self.categories:
+            mask = data["Record"]["type"] == category
             df = data["Record"][mask].copy()
-            df = self._special_cases(df, dtype)
-            self.df[dtype] = df
+            df = self._special_cases(df, category)
+            self.df[category] = df
         return self.dataframes
         
 
